@@ -1,36 +1,32 @@
-/**
-* @file dlgcontrollerlearning.cpp
-* @author Sean M. Pappalardo  spappalardo@mixxx.org
-* @date Thu 12 Apr 2012
-* @brief The controller mapping learning wizard
-*
-*/
+#include "controllers/dlgcontrollerlearning.h"
 
 #include <QCompleter>
 
-#include "controlobject.h"
-#include "controllers/dlgcontrollerlearning.h"
+#include "control/controlobject.h"
 #include "controllers/learningutils.h"
+#include "controllers/midi/midicontroller.h"
 #include "controllers/midi/midiutils.h"
-
+#include "moc_dlgcontrollerlearning.cpp"
+#include "util/versionstore.h"
 
 namespace {
 typedef QPair<QString, ConfigKey> NamedControl;
 bool namedControlComparator(const NamedControl& l1, const NamedControl& l2) {
     return l1.first < l2.first;
 }
-}
+} // namespace
 
-DlgControllerLearning::DlgControllerLearning(QWidget * parent,
-                                             Controller* controller)
+DlgControllerLearning::DlgControllerLearning(QWidget* parent,
+        Controller* controller,
+        ControlPickerMenu* pControlPickerMenu)
         : QDialog(parent),
           m_pController(controller),
-          m_pMidiController(NULL),
-          m_controlPickerMenu(this),
+          m_pControlPickerMenu(pControlPickerMenu),
           m_messagesLearned(false) {
     qRegisterMetaType<MidiInputMappings>("MidiInputMappings");
 
     setupUi(this);
+    labelDescription->setWordWrap(true);
     labelMappedTo->setText("");
 
     QString helpTitle(tr("Click anywhere in Mixxx or choose a control to learn"));
@@ -71,25 +67,47 @@ DlgControllerLearning::DlgControllerLearning(QWidget * parent,
     setAttribute(Qt::WA_DeleteOnClose);
     setWindowFlags(Qt::Tool | Qt::WindowStaysOnTopHint);
 
-    connect(&m_controlPickerMenu, SIGNAL(controlPicked(ConfigKey)),
-            this, SLOT(controlPicked(ConfigKey)));
+    connect(m_pControlPickerMenu,
+            &ControlPickerMenu::controlPicked,
+            this,
+            &DlgControllerLearning::controlPicked);
 
     comboBoxChosenControl->completer()->setCompletionMode(
         QCompleter::PopupCompletion);
     populateComboBox();
-    connect(comboBoxChosenControl, SIGNAL(currentIndexChanged(int)),
-            this, SLOT(comboboxIndexChanged(int)));
+    connect(comboBoxChosenControl,
+            QOverload<int>::of(&QComboBox::currentIndexChanged),
+            this,
+            &DlgControllerLearning::comboboxIndexChanged);
 
-    connect(pushButtonChooseControl, SIGNAL(clicked()), this, SLOT(showControlMenu()));
-    connect(pushButtonClose, SIGNAL(clicked()), this, SLOT(close()));
-    connect(pushButtonClose_2, SIGNAL(clicked()), this, SLOT(close()));
-    connect(pushButtonCancelLearn, SIGNAL(clicked()), this, SLOT(slotCancelLearn()));
-    connect(pushButtonRetry, SIGNAL(clicked()), this, SLOT(slotRetry()));
-    connect(pushButtonStartLearn, SIGNAL(clicked()), this, SLOT(slotStartLearningPressed()));
-    connect(pushButtonLearnAnother, SIGNAL(clicked()), this, SLOT(slotChooseControlPressed()));
+    connect(pushButtonChooseControl,
+            &QAbstractButton::clicked,
+            this,
+            &DlgControllerLearning::showControlMenu);
+    connect(pushButtonClose, &QAbstractButton::clicked, this, &DlgControllerLearning::close);
+    connect(pushButtonClose_2, &QAbstractButton::clicked, this, &DlgControllerLearning::close);
+    connect(pushButtonCancelLearn,
+            &QAbstractButton::clicked,
+            this,
+            &DlgControllerLearning::slotCancelLearn);
+    connect(pushButtonRetry, &QAbstractButton::clicked, this, &DlgControllerLearning::slotRetry);
+    connect(pushButtonStartLearn,
+            &QAbstractButton::clicked,
+            this,
+            &DlgControllerLearning::slotStartLearningPressed);
+    connect(pushButtonLearnAnother,
+            &QAbstractButton::clicked,
+            this,
+            &DlgControllerLearning::slotChooseControlPressed);
 #ifdef CONTROLLERLESSTESTING
-    connect(pushButtonFakeControl, SIGNAL(clicked()), this, SLOT(DEBUGFakeMidiMessage()));
-    connect(pushButtonFakeControl2, SIGNAL(clicked()), this, SLOT(DEBUGFakeMidiMessage2()));
+    connect(pushButtonFakeControl,
+            &QAbstractButton::clicked,
+            this,
+            &DlgControllerLearning::DEBUGFakeMidiMessage);
+    connect(pushButtonFakeControl2,
+            &QAbstractButton::clicked,
+            this,
+            &DlgControllerLearning::DEBUGFakeMidiMessage2);
 #else
     pushButtonFakeControl->hide();
     pushButtonFakeControl2->hide();
@@ -97,27 +115,36 @@ DlgControllerLearning::DlgControllerLearning(QWidget * parent,
 
     // We only want to listen to clicked() so we don't fire
     // slotMidiOptionsChanged when we change the checkboxes programmatically.
-    connect(midiOptionSwitchMode, SIGNAL(clicked()),
-            this, SLOT(slotMidiOptionsChanged()));
-    connect(midiOptionSoftTakeover, SIGNAL(clicked()),
-            this, SLOT(slotMidiOptionsChanged()));
-    connect(midiOptionInvert, SIGNAL(clicked()),
-            this, SLOT(slotMidiOptionsChanged()));
-    connect(midiOptionSelectKnob, SIGNAL(clicked()),
-            this, SLOT(slotMidiOptionsChanged()));
+    connect(midiOptionSwitchMode,
+            &QAbstractButton::clicked,
+            this,
+            &DlgControllerLearning::slotMidiOptionsChanged);
+    connect(midiOptionSoftTakeover,
+            &QAbstractButton::clicked,
+            this,
+            &DlgControllerLearning::slotMidiOptionsChanged);
+    connect(midiOptionInvert,
+            &QAbstractButton::clicked,
+            this,
+            &DlgControllerLearning::slotMidiOptionsChanged);
+    connect(midiOptionSelectKnob,
+            &QAbstractButton::clicked,
+            this,
+            &DlgControllerLearning::slotMidiOptionsChanged);
 
     slotChooseControlPressed();
 
     // Wait 1 second until we detect the control the user moved.
     m_lastMessageTimer.setInterval(1500);
     m_lastMessageTimer.setSingleShot(true);
-    connect(&m_lastMessageTimer, SIGNAL(timeout()),
-            this, SLOT(slotTimerExpired()));
+    connect(&m_lastMessageTimer, &QTimer::timeout, this, &DlgControllerLearning::slotTimerExpired);
 
     m_firstMessageTimer.setInterval(7000);
     m_firstMessageTimer.setSingleShot(true);
-    connect(&m_firstMessageTimer, SIGNAL(timeout()),
-            this, SLOT(slotFirstMessageTimeout()));
+    connect(&m_firstMessageTimer,
+            &QTimer::timeout,
+            this,
+            &DlgControllerLearning::slotFirstMessageTimeout);
 }
 
 void DlgControllerLearning::populateComboBox() {
@@ -125,13 +152,12 @@ void DlgControllerLearning::populateComboBox() {
     comboBoxChosenControl->clear();
     comboBoxChosenControl->addItem("", QVariant::fromValue(ConfigKey()));
     QList<NamedControl> sorted_controls;
-    foreach(ConfigKey key, m_controlPickerMenu.controlsAvailable())
-    {
+    foreach (ConfigKey key, m_pControlPickerMenu->controlsAvailable()) {
         sorted_controls.push_back(
-                NamedControl(m_controlPickerMenu.controlTitleForConfigKey(key),
-                             key));
+                NamedControl(m_pControlPickerMenu->controlTitleForConfigKey(key),
+                        key));
     }
-    qSort(sorted_controls.begin(), sorted_controls.end(),
+    std::sort(sorted_controls.begin(), sorted_controls.end(),
           namedControlComparator);
     foreach(NamedControl control, sorted_controls)
     {
@@ -143,7 +169,7 @@ void DlgControllerLearning::populateComboBox() {
 void DlgControllerLearning::resetWizard(bool keepCurrentControl) {
     m_firstMessageTimer.stop();
     m_lastMessageTimer.stop();
-    emit(clearTemporaryInputMappings());
+    emit clearTemporaryInputMappings();
 
     if (!keepCurrentControl) {
         m_currentControl = ConfigKey();
@@ -181,14 +207,51 @@ void DlgControllerLearning::slotChooseControlPressed() {
 void DlgControllerLearning::startListening() {
     // Start listening as soon as we're on this page -- that way advanced
     // users don't have to specifically click the "Learn" button.
-    // Get the underlying type of the Controller. This will call
-    // one of the visit() methods below immediately.
-    m_pController->accept(this);
-    emit(listenForClicks());
+    // Disconnect everything in both directions so we don't end up with duplicate connections
+    // after pressing the "Learn Another" button
+    MidiController* pMidiController = qobject_cast<MidiController*>(m_pController);
+    VERIFY_OR_DEBUG_ASSERT(pMidiController) {
+        // DlgControllerLearning should have only been created by DlgController if
+        // the Controller was a MidiController.
+        qWarning() << "Only MIDI controllers are supported by the learning wizard.";
+        return;
+    }
+    pMidiController->disconnect(this);
+    this->disconnect(pMidiController);
+
+    connect(pMidiController,
+            &MidiController::messageReceived,
+            this,
+            &DlgControllerLearning::slotMessageReceived);
+
+    connect(this,
+            &DlgControllerLearning::learnTemporaryInputMappings,
+            pMidiController,
+            &MidiController::learnTemporaryInputMappings);
+    connect(this,
+            &DlgControllerLearning::clearTemporaryInputMappings,
+            pMidiController,
+            &MidiController::clearTemporaryInputMappings);
+
+    connect(this,
+            &DlgControllerLearning::commitTemporaryInputMappings,
+            pMidiController,
+            &MidiController::commitTemporaryInputMappings);
+    connect(this,
+            &DlgControllerLearning::startLearning,
+            pMidiController,
+            &MidiController::startLearning);
+    connect(this,
+            &DlgControllerLearning::stopLearning,
+            pMidiController,
+            &MidiController::stopLearning);
+
+    emit startLearning();
+    emit listenForClicks();
 }
 
 void DlgControllerLearning::slotStartLearningPressed() {
-    if (m_currentControl.isNull()) {
+    if (!m_currentControl.isValid()) {
         return;
     }
     m_firstMessageTimer.start();
@@ -197,11 +260,11 @@ void DlgControllerLearning::slotStartLearningPressed() {
 
 #ifdef CONTROLLERLESSTESTING
 void DlgControllerLearning::DEBUGFakeMidiMessage() {
-    slotMessageReceived(MIDI_CC, 0x20, 0x41);
+    slotMessageReceived(MidiOpCode::ControlChange, 0x20, 0x41);
 }
 
 void DlgControllerLearning::DEBUGFakeMidiMessage2() {
-    slotMessageReceived(MIDI_CC, 0x20, 0x3F);
+    slotMessageReceived(MidiOpCode::ControlChange, 0x20, 0x3F);
 }
 #endif
 
@@ -209,7 +272,7 @@ void DlgControllerLearning::slotMessageReceived(unsigned char status,
                                                 unsigned char control,
                                                 unsigned char value) {
     // Ignore message since we don't have a control yet.
-    if (m_currentControl.isNull()) {
+    if (!m_currentControl.isValid()) {
         return;
     }
 
@@ -218,9 +281,10 @@ void DlgControllerLearning::slotMessageReceived(unsigned char status,
         return;
     }
 
-    MidiKey key;
-    key.status = status;
-    key.control = control;
+    // NOTE(rryan): We intend to use MidiKey(status, control) here rather than
+    // setting fields individually since we will use the MidiKey with an input
+    // mapping. See Issue #8432
+    MidiKey key(status, control);
 
     // Ignore all standard MIDI System Real-Time Messages because they
     // are continuously sent and prevent mapping of the pressed key.
@@ -250,11 +314,11 @@ void DlgControllerLearning::slotMessageReceived(unsigned char status,
     // We got a message, so we can cancel the taking-too-long timeout.
     m_firstMessageTimer.stop();
 
-    // Unless this is a MIDI_CC and the progress bar is full, restart the
+    // Unless this is a MidiOpCode::ControlChange and the progress bar is full, restart the
     // timer.  That way the user won't just push buttons forever and wonder
     // why the wizard never advances.
-    unsigned char opCode = MidiUtils::opCodeFromStatus(status);
-    if (opCode != MIDI_CC || progressBarWiggleFeedback->value() != 10) {
+    MidiOpCode opCode = MidiUtils::opCodeFromStatus(status);
+    if (opCode != MidiOpCode::ControlChange || progressBarWiggleFeedback->value() != 10) {
         m_lastMessageTimer.start();
     }
 }
@@ -295,33 +359,39 @@ void DlgControllerLearning::slotTimerExpired() {
     m_messagesLearned = true;
     m_mappings = mappings;
     pushButtonRetry->setEnabled(true);
-    emit(learnTemporaryInputMappings(m_mappings));
+    emit learnTemporaryInputMappings(m_mappings);
 
     QString midiControl = "";
     bool first = true;
     foreach (const MidiInputMapping& mapping, m_mappings) {
-        unsigned char opCode = MidiUtils::opCodeFromStatus(mapping.key.status);
+        MidiOpCode opCode = MidiUtils::opCodeFromStatus(mapping.key.status);
         bool twoBytes = MidiUtils::isMessageTwoBytes(opCode);
-        QString mappingStr = twoBytes ? QString("Status: 0x%1 Control: 0x%2 Options: 0x%03")
-                .arg(QString::number(mapping.key.status, 16).toUpper(),
-                     QString::number(mapping.key.control, 16).toUpper()
-                     .rightJustified(2, '0'),
-                     QString::number(mapping.options.all, 16).toUpper()
-                     .rightJustified(2, '0')) :
-                QString("0x%1 0x%2")
-                .arg(QString::number(mapping.key.status, 16).toUpper(),
-                     QString::number(mapping.options.all, 16).toUpper()
-                     .rightJustified(2, '0'));
+        QString mappingStr = twoBytes
+                ? QString("Status: 0x%1 Control: 0x%2 Options: 0x%03")
+                          .arg(QString::number(mapping.key.status, 16)
+                                          .toUpper(),
+                                  QString::number(mapping.key.control, 16)
+                                          .toUpper()
+                                          .rightJustified(2, '0'),
+                                  QString::number(mapping.options, 16)
+                                          .toUpper()
+                                          .rightJustified(2, '0'))
+                : QString("0x%1 0x%2")
+                          .arg(QString::number(mapping.key.status, 16)
+                                          .toUpper(),
+                                  QString::number(mapping.options, 16)
+                                          .toUpper()
+                                          .rightJustified(2, '0'));
 
         // Set the debug string and "Advanced MIDI Options" group using the
         // first mapping.
         if (first) {
             midiControl = mappingStr;
             MidiOptions options = mapping.options;
-            midiOptionInvert->setChecked(options.invert);
-            midiOptionSelectKnob->setChecked(options.selectknob);
-            midiOptionSoftTakeover->setChecked(options.soft_takeover);
-            midiOptionSwitchMode->setChecked(options.sw);
+            midiOptionInvert->setChecked(options.testFlag(MidiOption::Invert));
+            midiOptionSelectKnob->setChecked(options.testFlag(MidiOption::SelectKnob));
+            midiOptionSoftTakeover->setChecked(options.testFlag(MidiOption::SoftTakeover));
+            midiOptionSwitchMode->setChecked(options.testFlag(MidiOption::Switch));
             first = false;
         }
 
@@ -348,56 +418,25 @@ void DlgControllerLearning::slotMidiOptionsChanged() {
         return;
     }
 
-    emit(clearTemporaryInputMappings());
+    emit clearTemporaryInputMappings();
 
     // Go over every mapping and set its MIDI options to match the user's
     // choices.
     for (MidiInputMappings::iterator it = m_mappings.begin();
          it != m_mappings.end(); ++it) {
         MidiOptions& options = it->options;
-        options.sw = midiOptionSwitchMode->isChecked();
-        options.soft_takeover = midiOptionSoftTakeover->isChecked();
-        options.invert = midiOptionInvert->isChecked();
-        options.selectknob = midiOptionSelectKnob->isChecked();
+        options.setFlag(MidiOption::Switch, midiOptionSwitchMode->isChecked());
+        options.setFlag(MidiOption::SoftTakeover, midiOptionSoftTakeover->isChecked());
+        options.setFlag(MidiOption::Invert, midiOptionInvert->isChecked());
+        options.setFlag(MidiOption::SelectKnob, midiOptionSelectKnob->isChecked());
     }
 
-    emit(learnTemporaryInputMappings(m_mappings));
+    emit learnTemporaryInputMappings(m_mappings);
 }
 
 void DlgControllerLearning::commitMapping() {
-    emit(commitTemporaryInputMappings());
-    emit(inputMappingsLearned(m_mappings));
-}
-
-void DlgControllerLearning::visit(MidiController* pMidiController) {
-    m_pMidiController = pMidiController;
-
-    connect(m_pMidiController, SIGNAL(messageReceived(unsigned char, unsigned char, unsigned char)),
-            this, SLOT(slotMessageReceived(unsigned char, unsigned char, unsigned char)));
-
-    connect(this, SIGNAL(learnTemporaryInputMappings(MidiInputMappings)),
-            m_pMidiController, SLOT(learnTemporaryInputMappings(MidiInputMappings)));
-    connect(this, SIGNAL(clearTemporaryInputMappings()),
-            m_pMidiController, SLOT(clearTemporaryInputMappings()));
-
-    connect(this, SIGNAL(commitTemporaryInputMappings()),
-            m_pMidiController, SLOT(commitTemporaryInputMappings()));
-    connect(this, SIGNAL(startLearning()),
-            m_pMidiController, SLOT(startLearning()));
-    connect(this, SIGNAL(stopLearning()),
-            m_pMidiController, SLOT(stopLearning()));
-
-    emit(startLearning());
-}
-
-void DlgControllerLearning::visit(HidController* pHidController) {
-    qDebug() << "ERROR: DlgControllerLearning does not support HID devices.";
-    Q_UNUSED(pHidController);
-}
-
-void DlgControllerLearning::visit(BulkController* pBulkController) {
-    qDebug() << "ERROR: DlgControllerLearning does not support Bulk devices.";
-    Q_UNUSED(pBulkController);
+    emit commitTemporaryInputMappings();
+    emit inputMappingsLearned(m_mappings);
 }
 
 DlgControllerLearning::~DlgControllerLearning() {
@@ -409,17 +448,17 @@ DlgControllerLearning::~DlgControllerLearning() {
     }
 
     //If there was any ongoing learning, cancel it (benign if there wasn't).
-    emit(stopLearning());
-    emit(stopListeningForClicks());
+    emit stopLearning();
+    emit stopListeningForClicks();
 }
 
 void DlgControllerLearning::showControlMenu() {
-    m_controlPickerMenu.exec(pushButtonChooseControl->mapToGlobal(QPoint(0,0)));
+    m_pControlPickerMenu->exec(pushButtonChooseControl->mapToGlobal(QPoint(0, 0)));
 }
 
 void DlgControllerLearning::loadControl(const ConfigKey& key,
-                                        QString title,
-                                        QString description) {
+        const QString& title,
+        QString description) {
     // If we have learned a mapping and the user picked a new control then we
     // should tell the controller to commit the existing ones.
     if (m_messagesLearned) {
@@ -444,38 +483,39 @@ void DlgControllerLearning::loadControl(const ConfigKey& key,
     pushButtonStartLearn->setFocus();
 }
 
-void DlgControllerLearning::controlPicked(ConfigKey control) {
-    QString title = m_controlPickerMenu.controlTitleForConfigKey(control);
-    QString description = m_controlPickerMenu.descriptionForConfigKey(control);
+void DlgControllerLearning::controlPicked(const ConfigKey& control) {
+    QString title = m_pControlPickerMenu->controlTitleForConfigKey(control);
+    QString description = m_pControlPickerMenu->descriptionForConfigKey(control);
     loadControl(control, title, description);
 }
 
-void DlgControllerLearning::controlClicked(ControlObject* pControl) {
-    if (!pControl) {
-        return;
-    }
-
-    ConfigKey key = pControl->getKey();
-    if (!m_controlPickerMenu.controlExists(key)) {
+void DlgControllerLearning::controlClicked(const ConfigKey& controlKey) {
+    if (!m_pControlPickerMenu->controlExists(controlKey)) {
         qWarning() << "Mixxx UI element clicked for which there is no "
-                      "learnable control " << key.group << " " << key.item;
+                      "learnable control "
+                   << controlKey.group << " " << controlKey.item;
         QMessageBox::warning(
-                    this,
-                    tr("Mixxx"),
-                    tr("The control you clicked in Mixxx is not learnable.\n"
-                       "This could be because you are using an old skin"
-                       " and this control is no longer supported.\n"
-                       "\nYou tried to learn: %1,%2").arg(key.group, key.item),
-                    QMessageBox::Ok, QMessageBox::Ok);
+                this,
+                VersionStore::applicationName(),
+                tr("The control you clicked in Mixxx is not learnable.\n"
+                   "This could be because you are either using an old skin"
+                   " and this control is no longer supported, "
+                   "or you clicked a control that provides visual feedback"
+                   " and can only be mapped to outputs like LEDs via"
+                   " scripts.\n"
+                   "\nYou tried to learn: %1,%2")
+                        .arg(controlKey.group, controlKey.item),
+                QMessageBox::Ok,
+                QMessageBox::Ok);
         return;
     }
-    controlPicked(key);
+    controlPicked(controlKey);
 }
 
 void DlgControllerLearning::comboboxIndexChanged(int index) {
     ConfigKey control =
             comboBoxChosenControl->itemData(index).value<ConfigKey>();
-    if (control.isNull()) {
+    if (!control.isValid()) {
         labelDescription->setText(tr(""));
         pushButtonStartLearn->setDisabled(true);
         return;
